import sys
from io import BytesIO
from pathlib import Path
import srsly
import torch
from spacy.util import SimpleFrozenDict
from spacy.vectors import get_current_ops

from spacy_transformers.layers import hf_shim
from spacy_transformers.layers.hf_shim import HFShim
from spacy_transformers.data_classes import HFObjects
from spacy_transformers.util import make_tempdir

from transformers import AutoModel, AutoConfig, AutoTokenizer


def override_hf_shims_to_bytes():
    assert hf_shim.HFShim.to_bytes is not HFShimCustom.to_bytes
    origin = hf_shim.HFShim.to_bytes
    hf_shim.HFShim.to_bytes = HFShimCustom.to_bytes
    return origin

def recover_hf_shims_to_bytes(origin):
    assert hf_shim.HFShim.to_bytes is HFShimCustom.to_bytes
    hf_shim.HFShim.to_bytes = origin


def override_hf_shims_from_bytes():
    assert hf_shim.HFShim.from_bytes is not HFShimCustom.from_bytes
    origin = hf_shim.HFShim.from_bytes
    hf_shim.HFShim.from_bytes = HFShimCustom.from_bytes
    return origin

def recover_hf_shims_from_bytes(origin):
    assert hf_shim.HFShim.from_bytes is HFShimCustom.from_bytes
    hf_shim.HFShim.from_bytes = origin


class HFShimCustom(HFShim):

    def to_bytes(self):
        config = {}
        tok_dict = {}
        # weights_bytes = {}
        tok_cfg = {}
        trf_cfg = {}
        hf_model = self._hfmodel
        if hf_model.transformer is not None:
            tok_dict = {}
            config = hf_model.transformer.config.to_dict()
            tokenizer = hf_model.tokenizer
            with make_tempdir() as temp_dir:
                if hasattr(tokenizer, "vocab_file"):
                    vocab_file_name = tokenizer.vocab_files_names["vocab_file"]
                    vocab_file_path = str((temp_dir / vocab_file_name).absolute())
                    with open(vocab_file_path, "wb") as fileh:
                        fileh.write(hf_model.vocab_file_contents)
                    tokenizer.vocab_file = vocab_file_path
                tokenizer.save_pretrained(str(temp_dir.absolute()))
                for x in temp_dir.glob("**/*"):
                    if x.is_file():
                        tok_dict[x.name] = x.read_bytes()
            filelike = BytesIO()
            torch.save(self._model.state_dict(), filelike)
            filelike.seek(0)
            # weights_bytes = filelike.getvalue()
        else:
            tok_cfg = hf_model._init_tokenizer_config
            trf_cfg = hf_model._init_transformer_config
        msg = {
            "config": config,
            # "state": weights_bytes,
            "tokenizer": tok_dict,
            "_init_tokenizer_config": tok_cfg,
            "_init_transformer_config": trf_cfg,
        }
        return srsly.msgpack_dumps(msg)

    def from_bytes(self, bytes_data):
        msg = srsly.msgpack_loads(bytes_data)
        config_dict = msg["config"]
        tok_dict = msg["tokenizer"]
        if config_dict:
            with make_tempdir() as temp_dir:
                config_file = temp_dir / "config.json"
                srsly.write_json(config_file, config_dict)
                config = AutoConfig.from_pretrained(config_file)
                for x, x_bytes in tok_dict.items():
                    Path(temp_dir / x).write_bytes(x_bytes)
                tokenizer = None
                try:
                    tokenizer = AutoTokenizer.from_pretrained(str(temp_dir.absolute()))
                except (ValueError, OSError):
                    pass
                if tokenizer is None:
                    tok_config = srsly.read_json(str((temp_dir / "tokenizer_config.json").absolute()))
                    tokenizer_class_name = tok_config["tokenizer_class"].split(".")
                    if tokenizer_class_name == ["ElectraSudachipyTokenizer"]:
                        from sudachitra.tokenization_electra_sudachipy import ElectraSudachipyTokenizer as tokenizer_class
                        tokenizer = tokenizer_class(vocab_file=str((temp_dir / "vocab.txt").absolute()), **tok_config)
                    else:
                        from importlib import import_module
                        tokenizer_module = import_module(".".join(tokenizer_class_name[:-1]))
                        tokenizer_class = getattr(tokenizer_module, tokenizer_class_name[-1])

                vocab_file_contents = None
                if hasattr(tokenizer, "vocab_file"):
                    vocab_file_name = tokenizer.vocab_files_names["vocab_file"]
                    vocab_file_path = str((temp_dir / vocab_file_name).absolute())
                    with open(vocab_file_path, "rb") as fileh:
                        vocab_file_contents = fileh.read()

            ops = get_current_ops()
            if ops.device_type == "cpu":
                map_location = "cpu"
            else:  # pragma: no cover
                device_id = torch.cuda.current_device()
                map_location = f"cuda:{device_id}"

            if "state" in msg:
                transformer = AutoModel.from_config(config)
                filelike = BytesIO(msg["state"])
                filelike.seek(0)
                transformer.load_state_dict(torch.load(filelike, map_location=map_location))
            else:
                try:
                    transformer = AutoModel.from_pretrained(config._name_or_path, local_files_only=True)
                except OSError as e2:
                    print("trying to download model from huggingface hub:", config._name_or_path, "...", file=sys.stderr)
                    transformer = AutoModel.from_pretrained(config._name_or_path)
                    print("succeded", file=sys.stderr)

            transformer.to(map_location)
            self._model = transformer
            self._hfmodel = HFObjects(
                tokenizer,
                transformer,
                vocab_file_contents,
                SimpleFrozenDict(),
                SimpleFrozenDict(),
            )
        else:
            self._hfmodel = HFObjects(
                None,
                None,
                None,
                msg["_init_tokenizer_config"],
                msg["_init_transformer_config"],
            )
        return self



from setuptools import setup, find_packages


setup(
    author="Megagon Labs, Tokyo.",
    author_email="ginza@megagon.ai",
    description="ginza-transformers",
    entry_points={
        "spacy_factories": [
            "transformer_custom = ginza_transformers.pipeline_component:make_transformer_custom",
        ],
    },
    install_requires=[
        "spacy-transformers>=1.1.2,<1.2.0",
    ],
    license="MIT",
    name="ginza-transformers",
    packages=find_packages(include=["ginza_transformers", "ginza_transformers.layers"]),
    url="https://github.com/megagonlabs/ginza-transformers",
    version='0.4.0',
)



from pathlib import Path
import sys
from typing import Union, Dict

from transformers import AutoConfig, AutoModel, AutoTokenizer

from spacy_transformers.layers import transformer_model
from spacy_transformers.data_classes import HFObjects

from thinc.api import get_current_ops, CupyOps


def override_huggingface_from_pretrained():
    assert transformer_model.huggingface_from_pretrained is not huggingface_from_pretrained_custom
    origin = transformer_model.huggingface_from_pretrained
    transformer_model.huggingface_from_pretrained = huggingface_from_pretrained_custom
    return origin

def recover_huggingface_from_pretrained(origin):
    assert transformer_model.huggingface_from_pretrained is huggingface_from_pretrained_custom
    transformer_model.huggingface_from_pretrained = origin


def huggingface_from_pretrained_custom(
    source: Union[Path, str], tok_config: Dict, trf_config: Dict
) -> HFObjects:
    """Create a Huggingface transformer model from pretrained weights. Will
    download the model if it is not already downloaded.

    source (Union[str, Path]): The name of the model or a path to it, such as
        'bert-base-cased'.
    tok_config (dict): Settings to pass to the tokenizer.
    trf_config (dict): Settings to pass to the transformer.
    """
    if hasattr(source, "absolute"):
        str_path = str(source.absolute())
    else:
        str_path = source
    try:
        tokenizer = AutoTokenizer.from_pretrained(str_path, **tok_config)
    except ValueError as e:
        if "tokenizer_class" not in tok_config:
            raise e
        tokenizer_class_name = tok_config["tokenizer_class"].split(".")
        from importlib import import_module
        tokenizer_module = import_module(".".join(tokenizer_class_name[:-1]))
        tokenizer_class = getattr(tokenizer_module, tokenizer_class_name[-1])
        tokenizer = tokenizer_class(vocab_file=str_path + "/vocab.txt", **tok_config)
    vocab_file_contents = None
    if hasattr(tokenizer, "vocab_file"):
        with open(tokenizer.vocab_file, "rb") as fileh:
            vocab_file_contents = fileh.read()

    try:
        trf_config["return_dict"] = True
        config = AutoConfig.from_pretrained(str_path, **trf_config)
        transformer = AutoModel.from_pretrained(str_path, config=config)
    except OSError as e:
        try:
            transformer = AutoModel.from_pretrained(str_path, local_files_only=True)
        except OSError as e2:
            model_name = str(source)
            print("trying to download model from huggingface hub:", model_name, "...", file=sys.stderr)
            transformer = AutoModel.from_pretrained(model_name)
            print("succeded", file=sys.stderr)
    ops = get_current_ops()
    if isinstance(ops, CupyOps):
        transformer.cuda()
    return HFObjects(tokenizer, transformer, vocab_file_contents)



from pathlib import Path
import sys
from typing import List, Callable, Iterable, Optional, Union

from spacy.language import Language
from spacy.tokens import Doc
from thinc.api import Model, Config

from spacy_transformers.data_classes import FullTransformerBatch
from spacy_transformers.pipeline_component import Transformer, DOC_EXT_ATTR
from spacy.training import Example

from .layers.hf_shim_custom import override_hf_shims_from_bytes, recover_hf_shims_from_bytes
from .layers.transformer_model import override_huggingface_from_pretrained, recover_huggingface_from_pretrained


DEFAULT_CONFIG_STR = """
[transformer_custom]
max_batch_items = 4096

[transformer_custom.set_extra_annotations]
@annotation_setters = "spacy-transformers.null_annotation_setter.v1"

[transformer_custom.model]
@architectures = "spacy-transformers.TransformerModel.v3"

[transformer_custom.model.get_spans]
@span_getters = "spacy-transformers.strided_spans.v1"
window = 128
stride = 96
"""

DEFAULT_CONFIG = Config().from_str(DEFAULT_CONFIG_STR)


@Language.factory(
    "transformer_custom",
    assigns=[f"doc._.{DOC_EXT_ATTR}"],
    default_config=DEFAULT_CONFIG["transformer_custom"],
)
def make_transformer_custom(
    nlp: Language,
    name: str,
    model: Model[List[Doc], FullTransformerBatch],
    set_extra_annotations: Callable[[List[Doc], FullTransformerBatch], None],
    max_batch_items: int,
):
    return TransformerCustom(
        nlp.vocab,
        model,
        set_extra_annotations,
        max_batch_items=max_batch_items,
        name=name,
    )


class TransformerCustom(Transformer):

    def initialize(
        self,
        get_examples: Callable[[], Iterable[Example]],
        *,
        nlp: Optional[Language] = None,
    ):
        origin = override_huggingface_from_pretrained()
        try:
            super().initialize(get_examples, nlp=nlp)
        finally:
            recover_huggingface_from_pretrained(origin)

    def from_disk(
        self, path: Union[str, Path], *, exclude: Iterable[str] = tuple()
    ) -> "TransformerCustom":
        origin = override_hf_shims_from_bytes()
        try:
            super().from_disk(path, exclude=exclude)
        finally:
            recover_hf_shims_from_bytes(origin)
        return self



import sys

import spacy

from .layers.hf_shim_custom import override_hf_shims_to_bytes, recover_hf_shims_to_bytes


def main():
    org_spacy_model_path = sys.argv[1]
    dst_spacy_model_path = sys.argv[2]
    transformers_model_name = sys.argv[3]

    nlp = spacy.load(org_spacy_model_path)
    transformer = nlp.get_pipe("transformer")
    for i, node in enumerate(transformer.model.walk()):
        if node.shims:
            break
    else:
        assert False
    node.shims[0]._hfmodel.transformer.config._name_or_path = transformers_model_name
    node.shims[0]._hfmodel.tokenizer.save_pretrained(transformers_model_name)
    node.shims[0]._hfmodel.transformer.save_pretrained(transformers_model_name)
    override_hf_shims_to_bytes()
    try:
        origin = nlp.to_disk(dst_spacy_model_path)
    finally:
        recover_hf_shims_to_bytes(origin)


if __name__ == "__main__":
    main()
